/*
 * GEMINI level editor in OpenGL
 *
 * This program was not optimized in any way, I just wrote it as a simple level
 * editor.
 *
 * I can't guarantee this prog to work on every system. I wrote it mainly for
 * my own use, and I give it out to you, who are able to use it.
 *
 * This prog was written in X, then ported directly to Windows.
 *
 * Not compiled, since I do not have any Windows-compilers :/ You'll have to do
 * this part yourself...
 */

#include <windows.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <GL\gl.h>
#include <GL\glu.h>
#include <GL\glaux.h>

#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")

typedef unsigned char	byte;
typedef unsigned short	word;

/* ************* *
 * Windows-stuff *
 * ************* */

#define WINDOW_NAME	("GEMINI - level editor")

int			key;
char		keys[65536];
int			button;

int			width,
			height;
HINSTANCE	globalInstance;
HWND		windowHandler;
MSG			windowMessage;
HDC			hDeviceContext;
HGLRC		hRenderContext;
float		deltaTime;
int			mx,
			my,
			winX,
			winY;

GLuint		textList;
GLYPHMETRICSFLOAT	gmf[96];

__int64		timeFreq;
float		timeRes;
__int64		timeOld,
			timeNew;
int			timeDelta;

void display_setupDepth() {
	int				iPixelFormat;
	static PIXELFORMATDESCRIPTOR	pfd = {
		sizeof(PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
		PFD_TYPE_RGBA,
		16, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,
		16,
		0, 0,
		PFD_MAIN_PLANE,
		0, 0, 0, 0 };

	iPixelFormat = ChoosePixelFormat(hDeviceContext, &pfd);
	SetPixelFormat(hDeviceContext, iPixelFormat, &pfd);
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {
	switch (uMsg) {
		case WM_CREATE:
			hDeviceContext = GetDC(hWnd);
			display_setupDepth();
			hRenderContext = wglCreateContext(hDeviceContext);
			wglMakeCurrent(hDeviceContext, hRenderContext);
		break;
		case WM_CLOSE:
			wglMakeCurrent(hDeviceContext, 0);
			wglDeleteContext(hRenderContext);
			PostQuitMessage(0);
		break;
		case WM_KEYDOWN:
			key = (char)wParam;
			keys[key] = 1;
		break;
		case WM_KEYUP:
			keys[wParam] = 0;
			key = 0;
		break;
		case WM_LBUTTONDOWN:
			button = 1;
		break;
		case WM_LBUTTONUP:
			button = 0;
		break;
		case WM_RBUTTONDOWN:
			button = 2;
		break;
		case WM_RBUTTONUP:
			button = 0;
		case WM_MOUSEMOVE:
			mx = ((int)lParam & 0xFFFF) + winX;
			my = ((int)lParam >> 16 & 0xFFFF) + winY;
		break;
		default:
		break;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

void display_close();

char display_create(char *title, int w, int h) {
	int			style;
	WNDCLASS	windowClass;
	RECT		windowRect;
	HFONT		font;

	style = WS_OVERLAPPED | WS_BORDER | WS_CAPTION | WS_VISIBLE;

	windowClass.style			= 0;
	windowClass.lpfnWndProc		= WndProc;
	windowClass.cbClsExtra		= 0;
	windowClass.cbWndExtra		= 0;
	windowClass.hInstance		= globalInstance;
	windowClass.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
	windowClass.hCursor			= LoadCursor(NULL, IDC_ARROW);
	windowClass.hbrBackground	= GetStockObject(BLACK_BRUSH);
	windowClass.lpszMenuName	= "MainMenu";
	windowClass.lpszClassName	= WINDOW_NAME;
	if (!RegisterClass(&windowClass)) return 0;

	windowRect.left = 0;
	windowRect.top = 0;
	windowRect.right = w;
	windowRect.bottom = h;
	AdjustWindowRect(&windowRect, style, FALSE);

	width = windowRect.right - windowRect.left;
	height = windowRect.bottom - windowRect.top;

	if (!(windowHandler = CreateWindow(WINDOW_NAME, title, style,
				        0, 0, width, height,
				        NULL, NULL, globalInstance, NULL))) {
		UnregisterClass(WINDOW_NAME, globalInstance);
		return 0;
	}

	GetClientRect(windowHandler, &windowRect);
	winX = windowRect.left;
	winY = windowRect.top;
	width = windowRect.right - windowRect.left;
	height = windowRect.bottom - windowRect.top;

	ShowWindow(windowHandler, SW_SHOW);
	UpdateWindow(windowHandler);

	QueryPerformanceFrequency((LARGE_INTEGER*)&timeFreq);
	timeRes = 1.0f / timeFreq;
	QueryPerformanceFrequency((LARGE_INTEGER*)&timeOld);

	textList = glGenLists(96);
	font = CreateFont(-12, 0, 0, 0, FW_BOLD, FALSE, FALSE, FALSE, ANSI_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY, FF_DONTCARE | DEFAULT_PITCH, "Arial");
	SelectObject(hDeviceContext, font);
	wglUseFontOutlines(hDeviceContext, 32, 95, textList, 0, 0.2f, WGL_FONT_POLYGONS, gmf);

	return 1;
}

void display_close() {
	glDeleteLists(textList, 96);
	ShowWindow(windowHandler, SW_SHOWNORMAL);
	DestroyWindow(windowHandler);
	UnregisterClass(WINDOW_NAME, globalInstance);
}

void display_ready() {
	glClear(GL_COLOR_BUFFER_BIT);
	PeekMessage(&windowMessage, windowHandler, 0, 0, PM_REMOVE);
}

void gl_set() {
	glEnable(GL_TEXTURE_2D);
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, width, 0, height, 0, 1);
	glMatrixMode(GL_MODELVIEW);
}

void display_update() {
	SwapBuffers(hDeviceContext);

	QueryPerformanceFrequency((LARGE_INTEGER*)&timeNew);
	timeDelta = (int)((timeNew - timeOld) * timeRes);
	timeOld = timeNew;

	TranslateMessage(&windowMessage);
	DispatchMessage(&windowMessage);
}

void print(float x, float y, const char *in, ...) {
	va_list	args;
	char	text[256];

	va_start(args, in);
	vsprintf(text, in, args);
	va_end(args);
	glLoadIdentity();
	glTranslatef(x, y, 0);
	glScalef(11, 11, 1);
	glListBase(textList - 32);
	glCallLists((GLsizei)strlen(text), GL_BYTE, text);
}

/* ****** *
 * Editor *
 * ****** */

#define BAR_WIDTH	96
#define NOT_HEIGHT	48
#define WIN_SIZE	650

#define MAX_LEVELS	4
#define MAX_OBJECTS	256

byte	levels[MAX_LEVELS][1024];
int		vis[1024],
		curLev,
		f2p,
		f3p,
		f4p,
		f11p,
		f12p,
		butp;
byte	curType,
		curTex,
		curHealth,
		value,
		mode,
		texColor;
byte	levelBits[] = {	0x00, 0x80, 0x88, 0x90, 0x98, 0x40 },
		objBits[] = {	0,
						0x80, 0x80, 0x80, 0x80,
						0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40, 0x40,
						0x10, 0x10, 0x10, 0x10,
						0x8, 0x8, 0x8,
						0, 0, 0, 0, 0, 0, 0, 0 },
		objValues[] = {	0,
						0, 1, 2, 3,
						0, 1, 2, 3, 4, 5, 6, 7,
						0, 1, 2, 3,
						1, 2, 4,
						0, 1, 2, 3, 4, 5, 6, 7 };

typedef struct {
	float	x, y;
	byte	dir;
} Player;
Player	players[MAX_LEVELS];

typedef struct {
	byte	used;
	byte	attr;
	float	x, y;
	byte	health;
} Object;
Object	objects[MAX_LEVELS][MAX_OBJECTS];

float	notify;
char	str[256];

float	r, g, b;

int		gotObjectType;

GLuint	tex;

FILE	*file;

void loadLevels() {
	if (!(file = fopen("save.lev", "rb"))) {
		notify = 1;
		strcpy(str, "No saved levels present.");
		return;
	}
	if (fread(levels, 1, MAX_LEVELS * 1024, file) != MAX_LEVELS * 1024) {
		fclose(file);
		notify = 1;
		strcpy(str, "Saved levels are corrupt.");
		return;
	}
	if (fread(players, 1, MAX_LEVELS * sizeof(Player), file) != MAX_LEVELS * sizeof(Player)) {
		fclose(file);
		notify = 1;
		strcpy(str, "Saved levels are corrupt.");
		return;
	}
	if (fread(objects, 1, MAX_LEVELS * MAX_OBJECTS * sizeof(Object), file) != MAX_LEVELS * MAX_OBJECTS * sizeof(Object)) {
		fclose(file);
		notify = 1;
		strcpy(str, "Saved levels are corrupt.");
		return;
	}
	fclose(file);
	notify = 1;
	strcpy(str, "Loaded levels successfully!");
}

void saveLevels() {
	if (!(file = fopen("save.lev", "wb"))) {
		notify = 1;
		strcpy(str, "Could not save levels.");
		return;
	}
	if (fwrite(levels, 1, MAX_LEVELS * 1024, file) != MAX_LEVELS * 1024) {
		fclose(file);
		notify = 1;
		strcpy(str, "Could not write to save-file.");
		return;
	}
	if (fwrite(players, 1, MAX_LEVELS * sizeof(Player), file) != MAX_LEVELS * sizeof(Player)) {
		fclose(file);
		notify = 1;
		strcpy(str, "Could not write to save-file.");
		return;
	}
	if (fwrite(objects, 1, MAX_LEVELS * MAX_OBJECTS * sizeof(Object), file) != MAX_LEVELS * MAX_OBJECTS * sizeof(Object)) {
		fclose(file);
		notify = 1;
		strcpy(str, "Could not write to save-file.");
		return;
	}
	fclose(file);
	notify = 1;
	strcpy(str, "Saved levels successfully!");
}

void exportLevels() {
	int	i, j, k,
		buffer[65536],
		origPos,
		bufPos,
		searchPos,
		ofs,
		pos, len;

	if (!(file = fopen("levels.inc", "wb"))) {
		notify = 1;
		strcpy(str, "Could not export levels.");
		return;
	}

	fprintf(file, ";\n; Levels file\n;\n; Generated by GEMINI level editor\n;\n\n");

	for (i = 0; i < MAX_LEVELS; i++) {
		fprintf(file, "\nlevel%d:", i);
		/* Compress level */
		buffer[0] = levels[i][0];
		buffer[1] = levels[i][1];
		origPos = 2;
		bufPos = 2;
		do {
			searchPos = origPos - 129;
			if (searchPos < 0) searchPos = 0;

			/* Find match */
			len = origPos - searchPos;
			if (origPos + len > 1023) len = 1024 - origPos;
			do {
				ofs = 0;
				do {
					if (!memcmp(levels[i] + searchPos + ofs,
								levels[i] + origPos, len)) {
						pos = origPos - (searchPos + ofs);
						goto MATCH_FOUND;
					}
					ofs++;
				} while (ofs + len < 129 && ofs + len < origPos);
				len--;
			} while (len >= 2);
			pos = 0;
			len = 0;
MATCH_FOUND:

			if (len >= 2) {
				buffer[bufPos] = (pos - 2) | 0x80;
				bufPos++;
				buffer[bufPos] = len;
				bufPos++;
				origPos += len;
			}
			else {
				buffer[bufPos] = 0;
				bufPos++;
				buffer[bufPos] = levels[i][origPos];
				bufPos++;
				origPos++;
			}
		} while (origPos < 1024);

		/* Write level-size */
		fprintf(file, "\n\t.dw\t%d\n", bufPos);

		/* Write level data */
		for (j = 0; j < bufPos; j++) {
			if (!(j & 15)) fprintf(file, "\t.db\t");
			fprintf(file, "$%02X", buffer[j]);
			if ((j & 15) < 15 && j < bufPos - 1) fprintf(file, ",");
			else fprintf(file, "\n");
		}

		/* Write player data */

		fprintf(file, "\t.dw\t$%04X, $%04X\n", (int)(32 * 256 * players[i].x), (int)(32 * 256 * players[i].y));
		fprintf(file, "\t.db\t$%02X\n", players[i].dir);

		/* Write object data */

		k = 0;
		for (j = 0; j < MAX_OBJECTS; j++) if (objects[i][j].used) k++;
		fprintf(file, "\t.db\t%d\n", k);

		for (j = 0; j < MAX_OBJECTS; j++) {
			if (!objects[i][j].used) continue;
			fprintf(file, "\t.db\t$%02X\n\t.dw\t0\n", objects[i][j].attr);
			if ((objects[i][j].attr & 0x80)) {
				fprintf(file, "\t.dw\t$%04X, 0, $%04X, 0\n", (int)(32 * 256 * objects[i][j].x), (int)(32 * 256 * objects[i][j].y));
				fprintf(file, "\t.db\t(%d + ENEMY_DIE), 0\n", objects[i][j].health);
				fprintf(file, "\t.db\t0, 0\n");
			}
			else
				fprintf(file, "\t.dw\t$%04X, $%04X\n", (int)(32 * 256 * objects[i][j].x), (int)(32 * 256 * objects[i][j].y));
		}
/*		fprintf(file, "\t;Dummy data, please read devlp.txt for more info.\n");
		fprintf(file, "\t.dw\t0,0,0,0,0,0,0\n");*/
	}

	fclose(file);

	notify = 1;
	strcpy(str, "Exported levels successfully!");
}

int wallColor(byte in) {
/*	texColor = (((in & 0xF) - 1) << 3);*/
	texColor = 0;
	if ((in & 0x98) == 0x98) return 0x8000EE | texColor | texColor << 8;
	if ((in & 0x90) == 0x90) return 0x0080EE | texColor | texColor << 8;
	if ((in & 0x88) == 0x88) return 0x0000EE | texColor | texColor << 8;
	if ((in & 0x80)) return 0x00EE00 | texColor | texColor << 16;
	if ((in & 0x40)) return 0x00EEEE | texColor << 8;
	if (in) return 0xEE8000 | texColor;
	return 0;
}

void objColor(byte in) {
	if ((in & 0x80)) {
		/* Enemy */
		r = 1;
		g = 0.6;
		b = 0.3;
		return;
	}
	if ((in & 0x40)) {
		/* Grounded object */
		r = 0.75;
		g = 0.75;
		b = 0.75;
		return;
	}
	if ((in & 0x10)) {
		/* Weapon */
		r = 1;
		g = 1;
		b = 0;
		return;
	}
	if ((in & 0x8)) {
		/* Key */
		r = 0;
		g = 1;
		b = 1;
		return;
	}
	/* Boost */
	r = 0.3;
	g = 0.6;
	b = 1;
}

#define RETURN(x)	{ gotObjectType = x; return; }
void getObjectType(byte attr) {
	if ((attr & 0x80)) RETURN(attr & 0x7F);
	if ((attr & 0x40)) RETURN((attr & 0x7) + 4);
	if ((attr & 0x10)) RETURN((attr & 0x7) + 12);
	if (attr == 0x9) RETURN(16);
	if (attr == 0xA) RETURN(17);
	if (attr == 0xB) RETURN(18);
	RETURN((attr & 0x7) + 19);
}
#undef RETURN

void drawDot(float x, float y) {
	glVertex3f(x - 2, y - 2, 0);
	glVertex3f(x + 2, y - 2, 0);
	glVertex3f(x + 2, y + 2, 0);
	glVertex3f(x - 2, y + 2, 0);
}

void editor() {
	float	i;
	int		j;
	float	x, y;

	/* Disk IO */

	if (keys[VK_F3] && !f3p) {
		f3p = 1;
		loadLevels();
	}
	if (!keys[VK_F3]) f3p = 0;
	if (keys[VK_F2] && !f2p) {
		f2p = 1;
		saveLevels();
	}
	if (!keys[VK_F2]) f2p = 0;
	if (keys[VK_F4] && !f4p) {
		f4p = 1;
		exportLevels();
	}
	if (!keys[VK_F4]) f4p = 0;

	/* Level stepping */

	if (keys[VK_F11] && !f11p) {
		f11p = 1;
		curLev--;
		if (curLev < 0) curLev = MAX_LEVELS - 1;
	}
	if (!keys[VK_F11]) f11p = 0;
	if (keys[VK_F12] && !f12p) {
		f12p = 1;
		curLev++;
		if (curLev >= MAX_LEVELS) curLev = 0;
	}
	if (!keys[VK_F12]) f12p = 0;

	/* Check mouse clicking */

	x = (float)mx / (width - BAR_WIDTH);
	y = (float)my / (height - NOT_HEIGHT);

	if (!button) butp = 0;

	if (button && mx >= 0 && my >= 0 && mx < width - BAR_WIDTH && my < height - NOT_HEIGHT) {
		if (!mode) {
			if (button == 1) {
				value = levelBits[curType];
				if (curType == 0 || curType == 5) value |= curTex;
				else value |= (curTex - 1);
			}
			else value = 0;
			levels[curLev][(int)(32 * x) + (int)(32 * y) * 32] = value;
		}
		if (mode) {
			if (!curType) {
				players[curLev].x = x;
				players[curLev].y = y;
				players[curLev].dir = 0;
			}
			else {
				if (button == 1) {
					if (!butp) {
						butp = 1;
						j = 0;
						while (objects[curLev][j].used) j++;
						objects[curLev][j].used = 1;
						objects[curLev][j].attr = objBits[curType] | objValues[curType];
						objects[curLev][j].x = x;
						objects[curLev][j].y = y;
						objects[curLev][j].health = curHealth;
					}
				}
				if (button > 1) {
					for (j = 0; j < MAX_OBJECTS; j++)
						if (objects[curLev][j].used) {
							if (x > objects[curLev][j].x - 2.0f / width &&
								x < objects[curLev][j].x + 2.0f / width &&
								y > objects[curLev][j].y - 2.0f / height &&
								y < objects[curLev][j].y + 2.0f / height)
								objects[curLev][j].used = 0;
						}
				}
			}
		}
	}
	if (button && mx >= width - BAR_WIDTH && mx < width && my >= 0 && my < height - NOT_HEIGHT) {
		if (my < 13);
		else if (my < 26) {
			mode = 0;
			curType = 0;
		}
		else if (my < 39) {
			mode = 0;
			curType = 1;
		}
		else if (my < 52) {
			mode = 0;
			curType = 2;
		}
		else if (my < 65) {
			mode = 0;
			curType = 3;
		}
		else if (my < 78) {
			mode = 0;
			curType = 4;
		}
		else if (my < 91) {
			mode = 0;
			curType = 5;
		}
		else if (my < 286) {
			mode = 0;
			curTex = (my - 91) / 13 + 1;
			if (curType > 0 && curType < 5 && curTex > 8) curTex = 8;
			if (curType > 5) curType = 5;
		}
		else if (my < 299) {
			mode = 1;
			curType = 0;
		}
		else if (my < 312) {
			mode = 1;
			curType = 1;
			curHealth = 3;
		}
		else if (my < 325) {
			mode = 1;
			curType = 2;
			curHealth = 10;
		}
		else if (my < 338) {
			mode = 1;
			curType = 3;
			curHealth = 18;
		}
		else if (my < 351) {
			mode = 1;
			curType = 4;
			curHealth = 24;
		}
		else if (my < 650) {
			mode = 1;
			curType = (my - 351) / 13 + 5;
		}
	}

	for (j = 0; j < 1024; j++) vis[j] = wallColor(levels[curLev][j]);

	glLoadIdentity();

	glBindTexture(GL_TEXTURE_2D, tex);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, 32, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, vis);
	glBegin(GL_TRIANGLE_STRIP);
		glTexCoord2f(0, 0); glVertex3f(0, height, 0);
		glTexCoord2f(1, 0); glVertex3f(width - BAR_WIDTH, height, 0);
		glTexCoord2f(0, 1); glVertex3f(0, NOT_HEIGHT, 0);
		glTexCoord2f(1, 1); glVertex3f(width - BAR_WIDTH, NOT_HEIGHT, 0);
	glEnd();

	glDisable(GL_TEXTURE_2D);

	glColor3f(0.3, 0.3, 0.3);
	glBegin(GL_LINES);
		for (i = 0; i <= width - BAR_WIDTH; i += (width - BAR_WIDTH) / 32.0f) {
			glVertex3f(i, NOT_HEIGHT, 0);
			glVertex3f(i, height, 0);
		}
		for (i = height; i >= NOT_HEIGHT; i -= (height - NOT_HEIGHT) / 32.0f) {
			glVertex3f(0, i, 0);
			glVertex3f(width - BAR_WIDTH, i, 0);
		}
	glEnd();

	glBegin(GL_QUADS);
		glColor3f(1, 1, 1);
		drawDot(players[curLev].x * (width - BAR_WIDTH), height - players[curLev].y * (height - NOT_HEIGHT) - 1);

		for (j = 0; j < MAX_OBJECTS; j++) if (objects[curLev][j].used) {
			objColor(objects[curLev][j].attr);
			glColor3f(r, g, b);
			drawDot(objects[curLev][j].x * (width - BAR_WIDTH), height - objects[curLev][j].y * (height - NOT_HEIGHT));
		}
	glEnd();

	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE, GL_ONE);

	if (!mode) {
		glColor3f(0.5, 0, 0.5);
		glBegin(GL_TRIANGLE_STRIP);
			glVertex3f(width - BAR_WIDTH + 1, height - 13 * curType - 28, 0);
			glVertex3f(width, height - 13 * curType - 28, 0);
			glVertex3f(width - BAR_WIDTH + 1, height - 13 * curType - 16, 0);
			glVertex3f(width, height - 13 * curType - 16, 0);
		glEnd();

		glColor3f(0.5, 0, 0.5);
		glBegin(GL_TRIANGLE_STRIP);
			glVertex3f(width - BAR_WIDTH + 1, height - 13 * curTex - 93, 0);
			glVertex3f(width, height - 13 * curTex - 93, 0);
			glVertex3f(width - BAR_WIDTH + 1, height - 13 * curTex - 81, 0);
			glVertex3f(width, height - 13 * curTex - 81, 0);
		glEnd();

		if (mx >= 0 && mx < width - BAR_WIDTH && my >= 0 && my < height - NOT_HEIGHT) {
			j = levels[curLev][(int)(32.0f * x) + (int)(32.0f * y) * 32];
			if ((j & 0x80)) j = (j & 7) + 1;
			else j &= 15;
			if (j) {
				glColor3f(0, 0.5, 0.5);
				glBegin(GL_TRIANGLE_STRIP);
					glVertex3f(width - BAR_WIDTH + 1, height - 13 * j - 93, 0);
					glVertex3f(width, height - 13 * j - 93, 0);
					glVertex3f(width - BAR_WIDTH + 1, height - 13 * j - 81, 0);
					glVertex3f(width, height - 13 * j - 81, 0);
				glEnd();
			}
		}
	}
	else {
		glColor3f(0.5, 0, 0.5);
		glBegin(GL_TRIANGLE_STRIP);
			glVertex3f(width - BAR_WIDTH + 1, height - 13 * curType - 301, 0);
			glVertex3f(width, height - 13 * curType - 301, 0);
			glVertex3f(width - BAR_WIDTH + 1, height - 13 * curType - 289, 0);
			glVertex3f(width, height - 13 * curType - 289, 0);
		glEnd();

		glColor3f(0, 0.5, 0.5);
		for (j = 0; j < MAX_OBJECTS; j++) {
			if (!objects[curLev][j].used) continue;
			if (x > objects[curLev][j].x - 2.0f / width &&
				x < objects[curLev][j].x + 2.0f / width &&
				y > objects[curLev][j].y - 2.0f / height &&
				y < objects[curLev][j].y + 2.0f / height) {
				getObjectType(objects[curLev][j].attr);
				glBegin(GL_TRIANGLE_STRIP);
					glVertex3f(width - BAR_WIDTH + 1, height - 13 * gotObjectType - 314, 0);
					glVertex3f(width, height - 13 * gotObjectType - 314, 0);
					glVertex3f(width - BAR_WIDTH + 1, height - 13 * gotObjectType - 302, 0);
					glVertex3f(width, height - 13 * gotObjectType - 302, 0);
				glEnd();
			}
		}
	}

	glDisable(GL_BLEND);

	glColor3f(1, 1, 1);
	print(width - BAR_WIDTH + 8, height - 13, "LEVEL...: %d", curLev);

	glColor3f(0, 0.5, 1);
	print(width - BAR_WIDTH + 8, height - 26, "* Wall");
	glColor3f(0, 1, 0);
	print(width - BAR_WIDTH + 8, height - 39, "* Open door");
	glColor3f(1, 0, 0);
	print(width - BAR_WIDTH + 8, height - 52, "* Locked 0");
	glColor3f(1, 0.5, 0);
	print(width - BAR_WIDTH + 8, height - 65, "* Locked 1");
	glColor3f(1, 0, 0.5);
	print(width - BAR_WIDTH + 8, height - 78, "* Locked 2");
	glColor3f(1, 1, 0);
	print(width - BAR_WIDTH + 8, height - 91, "* Pushable");

	glColor3f(1, 1, 1);
	for (j = 1; j < 16; j++)
		print(width - BAR_WIDTH + 8, height - 91 - 13 * j, "* Tex %d", j);

	glColor3f(1, 1, 1);
	print(width - BAR_WIDTH + 8, height - 299, "* Player");

	glColor3f(1, 0.6, 0.3);
	print(width - BAR_WIDTH + 8, height - 312, "* Handgunner");
	print(width - BAR_WIDTH + 8, height - 325, "* Shotgunner");
	print(width - BAR_WIDTH + 8, height - 338, "* Chaingunner");
	print(width - BAR_WIDTH + 8, height - 351, "* Plasma'er'");

	glColor3f(0.75, 0.75, 0.75);
	print(width - BAR_WIDTH + 8, height - 364, "* Exit");
	for (j = 1; j < 8; j++)
		print(width - BAR_WIDTH + 8, height - 364 - 13 * j, "* Grounded %d", j);

	glColor3f(1, 1, 0);
	print(width - BAR_WIDTH + 8, height - 468, "* Handgun");
	print(width - BAR_WIDTH + 8, height - 481, "* Shotgun");
	print(width - BAR_WIDTH + 8, height - 494, "* Chaingun");
	print(width - BAR_WIDTH + 8, height - 507, "* Plasma rifle");

	glColor3f(0, 1, 0);
	print(width - BAR_WIDTH + 8, height - 520, "* Key 0");
	glColor3f(0, 1, 0);
	print(width - BAR_WIDTH + 8, height - 533, "* Key 1");
	glColor3f(0, 1, 0);
	print(width - BAR_WIDTH + 8, height - 546, "* Key 2");

	glColor3f(0.3, 0.6, 1);
	print(width - BAR_WIDTH + 8, height - 559, "* Small health");
	print(width - BAR_WIDTH + 8, height - 572, "* Big health");
	print(width - BAR_WIDTH + 8, height - 585, "* Small armor");
	print(width - BAR_WIDTH + 8, height - 598, "* Big armor");
	print(width - BAR_WIDTH + 8, height - 611, "* Clip");
	print(width - BAR_WIDTH + 8, height - 624, "* Shells");
	print(width - BAR_WIDTH + 8, height - 637, "* Bullets");
	print(width - BAR_WIDTH + 8, height - 650, "* Cells");

	glColor3f(0, 1, 1);
	print(8, 8, "[  ESC - quit  |  F2 - save  |  F3 - load  |  F4 - export  |  F11 - Previous level  |  F12 - Next level ]");

	if (notify > 0) {
		glColor3f(1, 1, 0);
		print(16, NOT_HEIGHT - 20, str);
		notify -= 0.5 * deltaTime;
	}

	glEnable(GL_TEXTURE_2D);
	glColor3f(1, 1, 1);

	glFlush();
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow) {
	GLenum	error;
	int		i, j;

	globalInstance = hInstance;

	if (!display_create("GEMINI - level editor", WIN_SIZE + BAR_WIDTH, WIN_SIZE + NOT_HEIGHT)) return 0;

	gl_set();

	glGenTextures(1, &tex);

	glBindTexture(GL_TEXTURE_2D, tex);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	curLev = 0;

	for (j = 0; j < MAX_LEVELS; j++) {
		for (i = 0; i < 1024; i++) levels[j][i] = 0;
		for (i = 0; i < MAX_OBJECTS; i++) objects[j][i].used = 0;
	}

	curTex = 0x01;
	mode = 0;

	do {
		display_ready();

		editor();

		display_update();
		while ((error = glGetError()) != GL_NO_ERROR)
			printf("GL-ERROR!!! Cause: %s\n", gluErrorString(error));
	} while (!keys[VK_ESCAPE]);

	glDeleteTextures(1, &tex);

	display_close();

	return 0;
}
